# -*- coding: utf-8 -*-
##############################################################################
#
#    Author: Guewen Baconnier
#    Copyright 2012 Camptocamp SA
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################
from datetime import datetime, timedelta

from openerp.osv import orm, fields
from openerp.tools import DEFAULT_SERVER_DATETIME_FORMAT

from .queue import JobsQueue
from .session import Session
from .jobs import Job


class JobsStorageModel(orm.Model):
    """ Job status and result.
    """
    _name = 'jobs.storage'

    _log_access = False

    _columns = {
        'worker_id': fields.many2one('queue.worker', string='Worker',
                                     ondelete='set null'),
        'uuid': fields.char('UUID', readonly=True, select=True),
        'user_id': fields.integer('User ID'),
        'name': fields.char('Description', readonly=True),
        'func_string': fields.char('Task', readonly=True),
        'func': fields.text('Pickled Job Function', readonly=True),
        # TODO: use the constants from module .tasks
        'state': fields.selection([('queued', 'Queued'),
                                   ('started', 'Started'),
                                   ('failed', 'Failed'),
                                   ('done', 'Done')],
                                  string='State',
                                  readonly=True),
        'exc_info': fields.text('Traceback', readonly=True),
        'result': fields.text('Result', readonly=True),
        'date_created': fields.datetime('Created Date', readonly=True),
        'date_started': fields.datetime('Start Date', readonly=True),
        'date_enqueued': fields.datetime('Enqueue Time', readonly=True),
        'date_done': fields.datetime('Date Done', readonly=True),
        'only_after': fields.datetime('Execute only after'),
        }

    def requeue(self, cr, uid, ids, context=None):
        if isinstance(ids, (int, long)):
            ids = [ids]
        session = Session(cr, uid, self.pool, context=context)
        for dbjob in self.browse(cr, uid, ids, context=context):
            if dbjob.state == 'failed':
                job = Job(job_id=dbjob.uuid)
                job.refresh(session)
                JobsQueue.instance.enqueue_job(session, job)
        return True


class QueueWorker(orm.Model):
    _name = 'queue.worker'

    worker_timeout = 300  # seconds: 5 minutes

    _log_access = False
    _workers = {}

    _columns = {
        'uuid': fields.char('UUID', readonly=True, select=True),
        'date_start': fields.datetime('Start Date', readonly=True),
        'date_alive': fields.datetime('Last Alive Check', readonly=True),
        'job_ids': fields.one2many('jobs.storage', 'worker_id',
                                   string='Jobs', readonly=True),
        }

    def _notify_alive(self, cr, uid, worker, context=None):
        worker_id = self._id_from_uuid(cr, uid, worker.uuid, context=context)

        now_fmt = datetime.now().strftime(DEFAULT_SERVER_DATETIME_FORMAT)
        if not worker_id:
            self.create(cr, uid,
                        {'uuid': worker.uuid,
                         'date_start': now_fmt,
                         'date_alive': now_fmt},
                        context=context)
            self._workers[worker.uuid] = worker
        else:
            self.write(cr, uid, worker_id,
                       {'date_start': now_fmt}, context=context)

    def _purge_dead_workers(self, cr, uid, context=None):
        deadline = datetime.now() - timedelta(seconds=self.worker_timeout)
        dead_ids = self.search(cr, uid,
                               [('date_alive', '<',
                                 deadline.strftime(DEFAULT_SERVER_DATETIME_FORMAT))],
                               context=context)
        # it will set worker_id to null on jobs, freeing them for
        # another worker
        dead_workers = self.read(cr, uid, dead_ids, ['uuid'], context=context)
        for worker in dead_workers:
            # exists in self._workers only for the same process and pool
            if worker.uuid in self._workers:
                del self._workers[worker.uuid]
        self.unlink(cr, uid, dead_ids, context=context)

    def _id_from_uuid(self, cr, uid, uuid, context=None):
        worker_ids = self.search(cr, uid, [('uuid', '=', uuid)],
                                 context=context)
        assert len(worker_ids) == 1
        return worker_ids[0]

    def assign_jobs(self, cr, uid, max_jobs=None, context=None):
        for worker in self._workers.itervalues():
            self._assign_jobs(cr, uid, worker, max_jobs=max_jobs, context=context)
        return True

    def enqueue_jobs(self, cr, uid, context=None):
        for worker in self._workers.itervalues():
            self._enqueue_jobs(cr, uid, worker, context=context)
        return True

    def _assign_jobs(self, cr, uid, worker, max_jobs=None, context=None):
        sql = ("SELECT id FROM jobs_storage "
               "WHERE worker_id IS NULL "
               "AND state not in ('failed', 'done') ")
        if max_jobs is not None:
            sql += ' LIMIT %d' % max_jobs
        sql += ' FOR UPDATE'
        job_rows = cr.execute(sql)
        job_ids = [id for id, in job_rows]

        worker_id = self._id_from_uuid(cr, uid, worker.uuid, context=context)
        self.pool.get('jobs.storage').write(cr, uid, job_ids,
                                            {'state': 'pending',
                                             'worker_id': worker_id},
                                            context=context)

    def _enqueue_jobs(self, cr, uid, worker, context=None):
        """ Called by an ir.cron, add to the queue all the jobs not
        already queued"""
        worker_id = self._id_from_uuid(cr, uid, worker.uuid, context=context)
        worker = self.browse(cr, uid, worker_id, context=context)
        for job in worker.job_ids:
            session = Session(cr, uid, self.pool,
                              self._name, context=context)
            worker.enqueue_job_uuid(session, job.uuid)
