# -*- coding: utf-8 -*-
##############################################################################
#
#    Author: Guewen Baconnier
#    Copyright 2012 Camptocamp SA
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################

# TODO: enqueue all pending jobs on startup
# assign the jobs to 1 queue at start of the server ideally, we would
# rely on the process number (if process_count % process_number: queue
# all) or check if we can identify the thread worker...
# Warning: the gunicorn server are killed and restarted based on some
# limits (memory, requests count, ...).
# so the enqueue have to be done only on the first startup
# If we can't know if it is a real startup or not, so only the process
# with the crons should have a queue...

# TODO: stop of the server
# check if we have issues with the thread which cannot write the state
# 'done' on the job if the server is stopped meanwhile
# How to handle that? A reasoning can be to ask tasks to always check if
# they still have to be done (for the ones which have absolutely not to
# be run twice).

from StringIO import StringIO
import traceback
import logging
import threading
import time
from datetime import datetime
from uuid import uuid4

import openerp
from .jobs import Job, QUEUED, STARTED, DONE, FAILED
from .queue import JobsQueue
from .session import Session
from .exceptions import (NoSuchJobError,
                         NotReadableJobError,
                         FailedJobError,
                         RetryableJobError)

_logger = logging.getLogger(__name__)

WAIT_REGISTRY_TIME = 20  # seconds
WAIT_WHEN_ONLY_AFTER_JOBS = 10  # seconds


class Worker(threading.Thread):

    def __init__(self, db_name, queue=JobsQueue.instance):
        super(Worker, self).__init__()
        self.queue = queue
        self.db_name = db_name
        self.registry = openerp.pooler.get_pool(db_name)
        self.uuid = unicode(uuid4())

    def run_job(self, job):
        """ """
        db = openerp.sql_db.db_connect(self.db_name)
        cr = db.cursor()

        with Session(cr, openerp.SUPERUSER_ID, self.registry) as session:
            try:
                try:
                    job.refresh(session)
                    if job.only_after and job.only_after > datetime.now():
                        # The queue is sorted by 'only_after' date first
                        # so if we dequeued a job expected to be run in
                        # the future, we have no jobs to do right now!
                        self.queue.enqueue_job(session, job)
                        # Wait some time just to avoid to loop over
                        # the same 'future' jobs
                        _logger.debug('Wait %s seconds because the delayed '
                                      'jobs have been reached',
                                      WAIT_WHEN_ONLY_AFTER_JOBS)
                        time.sleep(WAIT_WHEN_ONLY_AFTER_JOBS)
                        return
                except NoSuchJobError:
                    # just skip and continue to the next
                    return
                except NotReadableJobError:
                    # will be put in failed by the enclosing try/except
                    _logger.debug('Cannot read: %s', job)
                    raise
                job.set_state(session, STARTED)
                _logger.debug('Starting: %s', job)
                result = job.perform(session)
                _logger.debug('Done: %s', job)
                job.set_state(session, DONE, result=result)
            except RetryableJobError:
                # TODO: implement the retryable errors:
                # retryable should be requeued with a only_after date
                raise NotImplementedError('RetryableJobError to implement')
            except (FailedJobError, Exception):  # XXX Exception?
                # TODO allow to pass a pipeline of exception
                # handlers (log errors, send by email, ...)
                buff = StringIO()
                traceback.print_exc(file=buff)
                _logger.error(buff.getvalue())
                # the session cursor may be in an bad state
                error_session = Session(
                        db.cursor(), openerp.SUPERUSER_ID, self.registry)
                with error_session:
                    job.set_state(error_session, FAILED,
                                  exc_info=buff.getvalue())
                raise

    def run(self):
        """ """
        while True:
            while (self.registry.ready and
                   'connectors.installed' in self.registry.models):
                job = self.queue.dequeue()
                try:
                    self.run_job(job)
                except:
                    continue

            # Wait the start of the registry. Note that for a DB without
            # connectors, it will loop undefinitely. But as soon as
            # we install the modules the thread will start working.
            time.sleep(WAIT_REGISTRY_TIME)

    def enqueue_job_uuid(self, session, uuid):
        job = Job(job_id=uuid)
        job.refresh(session)
        job.set_state(session, QUEUED)
        self.queue.enqueue_job(session, job)

class WorkerWatcher(threading.Thread):
    """ inform that the worker is still alive """

    def __init__(self, db_name, workers):
        super(WorkerWatcher, self).__init__()
        if not hasattr(workers, '__iter__'):
            workers = [workers]
        self.workers = workers
        self.db_name = db_name
        self.registry = openerp.pooler.get_pool(db_name)

    def run(self):
        while True:
            while (self.registry.ready and
                   'connectors.installed' in self.registry.models):
                self.check_alive()
                time.sleep(30)

    def check_alive(self):
        db = openerp.sql_db.db_connect(self.db_name)
        cr = db.cursor()
        with Session(cr, openerp.SUPERUSER_ID, self.registry) as session:
            self._notify_alive(session)
            session.commit()
            self._purge_dead_workers(session)
            session.commit()

    def _notify_alive(self, session):
        dbworker_obj = session.pool.get('queue.worker')
        for worker in self.workers:
            dbworker_obj._notify_alive(session.cr,
                                       session.uid,
                                       worker,
                                       context=session.context)

    def _purge_dead_workers(self, session):
        dbworker_obj = session.pool.get('queue.worker')
        for worker in self.workers:
            dbworker_obj._purge_dead_workers(session.cr,
                                             session.uid,
                                             context=session.context)


def start_service():
    registries = openerp.modules.registry.RegistryManager.registries
    for db_name, registry in registries.iteritems():
        worker = Worker(db_name)
        worker.daemon = True
        worker.start()

        watcher = WorkerWatcher(db_name, worker)
        watcher.daemon = True
        watcher.start()

start_service()
